import java.text.ParseException
import java.awt.Color
import java.text.NumberFormat
import org.apache.log4j.Logger
import org.apache.commons.lang.StringUtils
import com.atlassian.jira.issue.Issue
import com.google.common.collect.Multimap
import com.google.common.collect.Ordering
import com.google.common.collect.TreeMultimap
import org.jfree.chart.axis.DateTickMarkPosition
import org.jfree.chart.axis.NumberAxis
import org.jfree.chart.plot.ValueMarker
import org.jfree.chart.renderer.xy.ClusteredXYBarRenderer
import org.jfree.chart.renderer.xy.StandardXYBarPainter
import org.jfree.chart.renderer.xy.XYStepRenderer
import org.jfree.data.time.Day
import org.jfree.data.time.TimeSeries
import org.jfree.data.time.TimeSeriesCollection
import org.joda.time.Days
import org.joda.time.DateTime


if (!binding.variables.containsKey("searchRequest")) {
	searchRequest = null // set it explicitly to avoid "not a property" exceptions
}

burnDownChart = new BurnDownChartTool(issues, searchRequest, pdfContent, binding.variables.containsKey("sprintCustomFieldId") ? sprintCustomFieldId : null, binding.variables.containsKey("storyPointsCustomFieldId") ? storyPointsCustomFieldId : null, doneStatusNames)
filename = "Sprint ${title}.pdf".toString() // "filename" is a standard Java String, not a GString

/**
 * Velocity tool to generate Burn Down Charts, Velocity Trend Charts, and some other key agile metrics with the JIRA PDF View Plugin.
 */
public class BurnDownChartTool {
	def log = Logger.getLogger(this.getClass())

	def issues
	def searchRequest
	def pdfContentTool
	def sprintCustomFieldName
	def storyPointsCustomFieldName
	def doneStatusNames

	def now = new Date()

	def startDate
	def endDate
	def initialWork = 0
	def addedWork = 0 // added after start
	def doneWork = 0

	def addedSeries
	def doneSeries
	def remainingSeries
	def idealSeries

	def static calendar = Calendar.instance

	def static START_AND_ENDDATE_FORMAT = "d/M/yyyy"
	def static ADDED_COLOR = new Color(0xdc3912)
	def static DONE_COLOR = new Color(0x109618)
	def static REMAINING_COLOR = new Color(0xd04437)
	def static IDEAL_COLOR = new Color(0xcccccc)

	public BurnDownChartTool(issues, searchRequest, pdfContentTool, sprintCustomFieldId, storyPointsCustomFieldId, doneStatusNames) {
		this.issues = issues
		this.searchRequest = searchRequest
		this.pdfContentTool = pdfContentTool

		if(sprintCustomFieldId != null) {
			sprintCustomFieldName = "customfield_${sprintCustomFieldId}"
		}
		if(storyPointsCustomFieldId != null) {
			storyPointsCustomFieldName = "customfield_${storyPointsCustomFieldId}"
		}

		this.doneStatusNames = doneStatusNames

		init()
	}

	private def init() {
		// get start and end dates from Greenhopper
		if(sprintCustomFieldName != null) {
			def sprint = issues.get(0).getCustomFieldValue(sprintCustomFieldName)
			if(sprint != null) {
				if(!sprint.startDate.isEmpty() && (sprint.startDate.get(0) != null)) {
					startDate = trimToDay(sprint.startDate.get(0).toDate())
				}
				if(!sprint.endDate.isEmpty() && (sprint.endDate.get(0) != null)) {
					endDate = trimToDay(sprint.endDate.get(0).toDate())
				}
			}
		}

		// parse start and end dates from the filter description
		if(((startDate == null) || (endDate == null)) && (searchRequest != null)) {
			try {
				def tmp = StringUtils.trimToNull(StringUtils.substringBeforeLast(searchRequest.description, "-"))

				def startDateAsString = StringUtils.trimToNull(StringUtils.contains(tmp, " ") ? StringUtils.substringAfterLast(tmp, " ") : tmp)
				if(startDateAsString != null) {
					startDate = new Date().parse(START_AND_ENDDATE_FORMAT, startDateAsString)
				}

				def endDateAsString = StringUtils.trimToNull(StringUtils.substringAfterLast(searchRequest.description, "-"))
				if(endDateAsString != null) {
					endDate = new Date().parse(START_AND_ENDDATE_FORMAT, endDateAsString)
				}
			} catch(ParseException ex) {
				// catch silently, dates remain unparsed
			}
		}

		// calculate
		if(datesParsed()) {
			def preDate = new Day(startDate - 1)

			// calculate daily added and done
			def dailyAdded = TreeMultimap.create(Ordering.natural(), new IssueByCreationDateComparator())
			def dailyDone = TreeMultimap.create(Ordering.natural(), new IssueByCreationDateComparator())

			issues.each {
				dailyAdded.put(trimToDay(it.getCreated()), it)

				def doneDate = getIssueDoneDate(it)
				if(doneDate != null) {
					dailyDone.put(trimToDay(doneDate), it)
				}
			}

			// calculate initial (work added before startDate, exclusive) and added (work added between startDate and endDate, inclusive)
			dailyAdded.keySet().each {
				if(it.before(startDate)) {
					// accumulate if added before the sprint's time window
					initialWork += sumWork(dailyAdded.get(it))
				} else if(!it.after(endDate)) {
					// accumulate if added in the sprint's time window
					addedWork += sumWork(dailyAdded.get(it))
				}
			}
			log.debug("Initial work calculated")

			// calculate done (work done between startDate and now, inclusive)
			dailyDone.keySet().each {
				if(!it.before(startDate) && !it.after(now)) {
					// accumulate if done until now
					doneWork += sumWork(dailyDone.get(it))
				}
			}
			log.debug("Done work calculated")

			// calculate remaining (work not done, at each day's end)
			remainingSeries = new TimeSeries("remaining")
			def remaining = initialWork
			remainingSeries.add(preDate, initialWork) // extra data point that represents the moment when the sprint starts
			for (date in startDate..endDate) {
				if(!date.after(now)) {
					if(dailyAdded.containsKey(date)) {
						remaining += sumWork(dailyAdded.get(date))
					}
					if(dailyDone.containsKey(date)) {
						remaining -= sumWork(dailyDone.get(date))
					}

					remainingSeries.add(new Day(date), remaining)
				}
			}
			log.debug("Remaining work calculated")

			// calculate ideal
			idealSeries = new TimeSeries("ideal")
			idealSeries.add(preDate, initialWork)
			idealSeries.addOrUpdate(new Day(endDate), 0)

			// calculate added and done
			addedSeries = new TimeSeries("added")
			doneSeries = new TimeSeries("done")
			addedSeries.add(preDate, 0) // to scale identically with Burn Down
			doneSeries.add(preDate, 0)
			for (date in startDate..endDate) {
				addedSeries.add(new Day(date), sumWork(dailyAdded.get(date)))
				doneSeries.add(new Day(date), sumWork(dailyDone.get(date)))
			}
			log.debug("Added and done work calculated")
		}
	}

	def datesParsed() {
		return (startDate != null) && (endDate != null)
	}

	def isStarted() {
		return now.after(startDate)
	}

	def isFinished() {
		return now.after(endDate)
	}

	def getLength() {
		return 1 + Days.daysBetween(new DateTime(startDate), new DateTime(endDate)).days
	}

	def getIdealVelocity() {
		return getVelocity(getInitialWork(), getLength())
	}

	def getIdealVelocityFormatted() {
		return formatVelocity(getIdealVelocity())
	}

	def getGoneDays() {
		def goneDays = Days.daysBetween(new DateTime(startDate), new DateTime(now)).days
		def length = getLength()

		if(goneDays <= 0) {
			return 0
		}
		if(goneDays > length) {
			return length
		}

		return goneDays
	}

	def getActualVelocityFormatted() {
		return formatVelocity(getVelocity(getDoneWork(), getGoneDays()))
	}

	def getRemainingDays() {
		return length - goneDays
	}

	def getRemainingWork() {
		return initialWork + addedWork - doneWork
	}

	def getRequiredVelocityFormatted() {
		return formatVelocity(getVelocity(getRemainingWork(), getRemainingDays()))
	}

	TimeSeriesCollection getBurnDownDataset() {
		def dataset = new TimeSeriesCollection()
		dataset.addSeries(remainingSeries)
		dataset.addSeries(idealSeries)

		return dataset
	}

	def getBurnDownCustomizer() {
		return { chart ->
			chart.plot.renderer.setSeriesPaint(0, REMAINING_COLOR)
			chart.plot.renderer.setSeriesStroke(0, RenderingUtils.LINE_STROKE)
			chart.plot.renderer.setSeriesShapesVisible(0, true)
			chart.plot.renderer.setSeriesShape(0, RenderingUtils.CIRCLE_SHAPE)

			chart.plot.renderer.setSeriesPaint(1, IDEAL_COLOR)
			chart.plot.renderer.setSeriesStroke(1, RenderingUtils.MARKER_STROKE)

			chart.plot.rangeAxis.setLabel(units)
			chart.plot.rangeAxis.setStandardTickUnits(NumberAxis.createIntegerTickUnits())
		}
	}

	TimeSeriesCollection getVelocityDataset() {
		def dataset = new TimeSeriesCollection()
		dataset.addSeries(addedSeries)
		dataset.addSeries(doneSeries)

		return dataset
	}

	def getVelocityCustomizer() {
		return { chart ->
			def renderer = new ClusteredXYBarRenderer()

			renderer.barPainter = new StandardXYBarPainter()
			renderer.drawBarOutline = true
			renderer.shadowVisible = false
			renderer.margin = 0.25

			renderer.setSeriesPaint(0, ADDED_COLOR)
			renderer.setSeriesOutlinePaint(0, RenderingUtils.BACKGROUND_COLOR)

			renderer.setSeriesPaint(1, DONE_COLOR)
			renderer.setSeriesOutlinePaint(1, RenderingUtils.BACKGROUND_COLOR)

			def idealVelocityMarker = new ValueMarker(getIdealVelocity())
			idealVelocityMarker.stroke = RenderingUtils.MARKER_STROKE
			idealVelocityMarker.paint = IDEAL_COLOR

			chart.plot.renderer = renderer
			chart.plot.rangeAxis.setLabel(units)
			chart.plot.rangeAxis.lowerBound = 0
			chart.plot.rangeAxis.setStandardTickUnits(NumberAxis.createIntegerTickUnits())
			chart.plot.domainAxis.tickMarkPosition = DateTickMarkPosition.MIDDLE
			chart.plot.addRangeMarker(idealVelocityMarker)
		}
	}

	// - utils ----------------------------------------------------------------

	def getUnits() {
		return isStoryPointBased() ? "Story Points" : "Issues"
	}

	private boolean isStoryPointBased() {
		return (storyPointsCustomFieldName != null)
	}

	private int sumWork(def issues) {
		if(isStoryPointBased()) {
			// return story point sum
			def storyPoints = 0
			issues.each {
				def storyPoint = it.getCustomFieldValue(storyPointsCustomFieldName)
				if(storyPoint != null) {
					storyPoints += storyPoint
				}
			}

			return storyPoints
		}

		// return issue count
		return issues.size()
	}

	/**
	 * Override this to have a custom logic to decide the "done date".
	 * By default, it returns the latest date when the issue entered a status listed in <code>doneStatusNames</code>.
	 */
	private Date getIssueDoneDate(def issue) {
		def doneDate

		def changeHistories = pdfContentTool.changeHistoriesByIssue(issue)
		changeHistories.each {
			it.changeItemBeans.each {
			if((it.field == "status") && doneStatusNames.contains(it.toString)) {
					doneDate = it.created
				}
			}
		}

		return doneDate
	}

	private static getVelocity(def tasks, def days) {
		double velocity = ((double)tasks) / days
		return velocity
	}

	private static formatVelocity(def velocity) {
		return String.format("%.2f", velocity)
	}

	private static trimToDay(date) {
		calendar.setTime(date)

		calendar.set(Calendar.HOUR_OF_DAY, 0)
		calendar.set(Calendar.MINUTE, 0)
		calendar.set(Calendar.SECOND, 0)
		calendar.set(Calendar.MILLISECOND, 0)

		return calendar.time
	}

	private static class IssueByCreationDateComparator implements Comparator {
		int compare(issue1, issue2) {
			int c = issue1.getCreated().compareTo(issue2.getCreated())
			return (c != 0) ? c : issue1.key.compareTo(issue2.key)
		}
	}
}
