import com.atlassian.jira.issue.Issue
import com.google.common.base.Predicate
import com.google.common.collect.Collections2
import org.joda.time.DateMidnight
import org.joda.time.DateTime
import org.joda.time.Days


projectReport = new ProjectReportTool(issues, startDateCustomFieldId, endDateCustomFieldId)
filename = "${title} Status Report.pdf".toString() // "filename" is a standard Java String, not a GString

/**
 * Velocity tool to generate issue lists, Key Performance Metrics and status flags for project reports.
 * @see project-report-fo.vm as the actual template.
 */
public class ProjectReportTool {
	enum ProjectStatus { RED, YELLOW, GREEN }

	def issues
	def startDateCustomFieldName
	def endDateCustomFieldName

	def now = new Date()

	/* Reported period. */
	def startDate
	def endDate
	/* Next reported period. */
	def nextStartDate
	def nextEndDate

	def static PERIOD_LENGTH = 7

	public ProjectReportTool(issues, startDateCustomFieldId, endDateCustomFieldId) {
		this.issues = issues
		this.startDateCustomFieldName = "customfield_${startDateCustomFieldId}"
		this.endDateCustomFieldName = "customfield_${endDateCustomFieldId}"

		init()
	}

	private def init() {
		startDate = new DateMidnight().minusDays(PERIOD_LENGTH).toDate()
		endDate = new DateMidnight().toDate()

		nextStartDate = endDate
		nextEndDate = new DateMidnight().plusDays(PERIOD_LENGTH).toDate()
	}

	ProjectStatus getOverallStatus() {
		def total = getOverdueIssues().size() + getOverbudgetIssues().size() + getRiskIssues().size()
		return (total > 1) ? ProjectStatus.RED : ((total == 0) ? ProjectStatus.GREEN : ProjectStatus.YELLOW)
	}

	int getCompletionPercentage() {
		return (100*Collections2.filter(issues, new IssueCompletedPredicate()).size()) / issues.size()
	}

	def getCompletedIssues() {
		return Collections2.filter(issues, new IssueCompletedBetweenPredicate(from: startDate, to: endDate))
	}

	def getPlannedIssues() {
		return Collections2.filter(issues, new IssuePlannedBetweenPredicate(from: nextStartDate, to: nextEndDate))
	}

	def getOverdueIssues() {
		return Collections2.filter(issues, new IssueOverduePredicate(now: now))
	}

	def getOverbudgetIssues() {
		return Collections2.filter(issues, new IssueOverbudgetPredicate())
	}

	def getRiskIssues() {
		return Collections2.filter(issues, new IssueRiskPredicate())
	}

	def getIssueStartDate(issue) {
		return issue.getCustomFieldValue(startDateCustomFieldName)
	}

	def getIssueEndDate(issue) {
		return issue.getCustomFieldValue(endDateCustomFieldName)
	}

	def static daysBetweenDateAndNow(date) {
		return (date != null) ? Days.daysBetween(new DateTime(date), DateTime.now()).getDays() : -1
	}

	private class IssueCompletedPredicate implements Predicate {
		boolean apply(issue) {
			return (issue.resolutionDate != null)
		}
	}

	private class IssueCompletedBetweenPredicate implements Predicate {
		def from
		def to

		boolean apply(issue) {
			return (issue.resolutionDate != null) && !issue.resolutionDate.before(from) && !issue.resolutionDate.after(to)
		}
	}

	private class IssuePlannedBetweenPredicate implements Predicate {
		def from
		def to

		boolean apply(issue) {
			def start = getIssueStartDate(issue)
			def end = getIssueEndDate(issue)
			return (start != null) && (end != null) && !start.after(to) && !end.before(from)
		}
	}

	private class IssueOverduePredicate implements Predicate {
		def now

		boolean apply(issue) {
			def end = getIssueEndDate(issue)
			return (issue.resolutionDate == null) && (end != null) && end.before(now)
		}
	}

	private class IssueOverbudgetPredicate implements Predicate {
		boolean apply(issue) {
			return (issue.aggregateOriginalEstimate != null) && (issue.aggregateTimeSpent != null) && (issue.aggregateTimeSpent > issue.aggregateOriginalEstimate)
		}
	}

	private class IssueRiskPredicate implements Predicate {
		boolean apply(issue) {
			return (issue.resolutionDate == null) && ["Blocker", "Critical"].contains(issue.priorityObject.name)
		}
	}
}
