import java.awt.Color
import java.text.NumberFormat;
import java.text.SimpleDateFormat
import com.atlassian.jira.issue.Issue
import com.google.common.base.Predicate
import com.google.common.collect.ArrayListMultimap
import com.google.common.collect.Collections2
import com.google.common.collect.HashMultiset
import com.google.common.collect.Multimap
import com.google.common.collect.Ordering
import com.google.common.collect.TreeMultimap
import com.google.common.collect.TreeMultiset
import org.jfree.chart.axis.DateTickMarkPosition
import org.jfree.chart.plot.ValueMarker
import org.jfree.chart.renderer.xy.ClusteredXYBarRenderer
import org.jfree.chart.renderer.xy.StandardXYBarPainter
import org.jfree.data.category.CategoryDataset
import org.jfree.data.category.DefaultCategoryDataset
import org.jfree.data.general.PieDataset
import org.jfree.data.general.DefaultPieDataset
import org.jfree.data.time.Month
import org.jfree.data.time.TimeSeries
import org.jfree.data.time.TimeSeriesCollection
import org.apache.commons.lang.StringUtils
import org.apache.commons.lang.WordUtils


salesReport = new SalesReportTool(issues, leadValueCustomFieldId)
filename = "${title} - ${SalesReportTool.monthOnlyDateFormat.format(new Date())}.pdf".toString() // "filename" is a standard Java String, not a GString

/**
 * Velocity tool to generate sales reports with the JIRA PDF View Plugin.
 */
public class SalesReportTool {
	def issues

	def leadValueCustomFieldName

	def static currencySign = "USD"
	def static moneyFormat = NumberFormat.instance
	def static calendar = Calendar.instance
	def static monthOnlyDateFormat = new SimpleDateFormat("yyyy MMM")

	def static LOST_COLOR = new Color(0xdc3912)
	def static WON_COLOR = new Color(0x109618)
	def static OPEN_COLOR = new Color(0xff9900)

	public SalesReportTool(issues, leadValueCustomFieldId) {
		this.issues = issues

		leadValueCustomFieldName = "customfield_${leadValueCustomFieldId}"
	}

	/**
	 * Get the month to leads mapping, sorted by date increasing date (key) and increasing update date within the month (value).
	 */
	Multimap<Date, Issue> getMonthlyLeads(issues) {
		def monthlyLeads = TreeMultimap.create(Ordering.natural(), new IssueByUpdateDateComparator())

		issues.each {
			monthlyLeads.put(trimToMonth(it.updated), it)
		}

		return monthlyLeads
	}

	boolean isWonLead(issue) {
		return issue.statusObject.name.toLowerCase() == "won"
	}

	boolean isLostLead(issue) {
		return issue.statusObject.name.toLowerCase() == "lost"
	}

	String getLeadCurrency(issue) {
		return currencySign
	}

	long getLeadValue(issue) {
		def value = issue.getCustomFieldValue(leadValueCustomFieldName)
		return (value != null) ? value.longValue() : 0
	}

	/**
	 * Extracts country name as the last line of the Description field.
	 */
	String getLeadCountry(issue) {
		  def leadFullAddress = StringUtils.defaultIfBlank(issue.environment, "Unknown")
		return WordUtils.capitalize(StringUtils.trimToEmpty(StringUtils.substringAfterLast(StringUtils.trimToEmpty(StringUtils.chomp(leadFullAddress.toLowerCase())), "\n")))
	}

	int getLeadCount(issues) {
		return issues.size()
	}

	String getLeadPercentage(issues, allIssues) {
		return getPercentageString(getLeadCount(issues), allIssues.size())
	}

	long getLeadTotalValue(issues) {
		long totalValue = 0

		issues.each { totalValue += getLeadValue(it) }

		return totalValue
	}

	// - Data calculators ---

	TimeSeriesCollection getWonVsLostDataset() {
		def monthlyLeads = getMonthlyLeads(issues)

		def wonSeries = new TimeSeries("won")
		def lostSeries = new TimeSeries("lost")
		monthlyLeads.keySet().each {
			def month = new Month(it)
			def leads = monthlyLeads.get(it)
			wonSeries.add(month, getLeadTotalValue(filterIssuesByStatusName(leads, "won")))
			lostSeries.add(month, getLeadTotalValue(filterIssuesByStatusName(leads, "lost")))
		}

		def dataset = new TimeSeriesCollection()
		dataset.addSeries(wonSeries)
		dataset.addSeries(lostSeries)

		return dataset
	}

	def getWonVsLostCustomizer() {
		return { chart ->
			def renderer = new ClusteredXYBarRenderer()

			renderer.barPainter = new StandardXYBarPainter()
			renderer.drawBarOutline = true
			renderer.shadowVisible = false
			renderer.margin = 0.25

			renderer.setSeriesPaint(0, WON_COLOR)
			renderer.setSeriesOutlinePaint(0, RenderingUtils.BACKGROUND_COLOR)
			renderer.setSeriesStroke(0, RenderingUtils.LINE_STROKE)

			renderer.setSeriesPaint(1, LOST_COLOR)
			renderer.setSeriesOutlinePaint(1, RenderingUtils.BACKGROUND_COLOR)
			renderer.setSeriesStroke(1, RenderingUtils.LINE_STROKE)

			def marker = new ValueMarker(2500)
			marker.stroke = RenderingUtils.MARKER_STROKE
			marker.paint = RenderingUtils.MARKER_COLOR

			chart.plot.renderer = renderer
			chart.plot.domainAxis.tickMarkPosition = DateTickMarkPosition.MIDDLE
			chart.plot.addRangeMarker(marker)
		}
	}

	TimeSeriesCollection getNewVsClosedDataset() {
		def newByMonth = new TreeMultiset()
		def wonByMonth = new TreeMultiset()
		def lostByMonth = new TreeMultiset()
		issues.each {
			newByMonth.add(new Month(it.getCreated()))

			if(isWonLead(it)) {
				wonByMonth.add(new Month(it.getUpdated()))
			} else if(isLostLead(it)) {
				lostByMonth.add(new Month(it.getUpdated()))
			}
		}

		def newSeries = new TimeSeries("new")
		def wonSeries = new TimeSeries("won")
		def lostSeries = new TimeSeries("lost")
		for(month in newByMonth.elementSet()) {
			newSeries.add(month, newByMonth.count(month));
			wonSeries.add(month, wonByMonth.count(month));
			lostSeries.add(month, lostByMonth.count(month));
		}

		def dataset = new TimeSeriesCollection()
		dataset.addSeries(newSeries)
		dataset.addSeries(wonSeries)
		dataset.addSeries(lostSeries)

		return dataset
	}

	def getNewVsClosedCustomizer() {
		return { chart ->
			chart.plot.renderer.fillPaint = RenderingUtils.BACKGROUND_COLOR
			chart.plot.renderer.useFillPaint = true

			chart.plot.renderer.setSeriesPaint(0, OPEN_COLOR)
			chart.plot.renderer.setSeriesShape(0, RenderingUtils.CIRCLE_SHAPE)
			chart.plot.renderer.setSeriesShapesVisible(0, true)

			chart.plot.renderer.setSeriesPaint(1, WON_COLOR)
			chart.plot.renderer.setSeriesShape(1, RenderingUtils.CIRCLE_SHAPE)
			chart.plot.renderer.setSeriesShapesVisible(1, true)

			chart.plot.renderer.setSeriesPaint(2, LOST_COLOR)
			chart.plot.renderer.setSeriesShape(2, RenderingUtils.CIRCLE_SHAPE)
			chart.plot.renderer.setSeriesShapesVisible(2, true)
		}
	}

	PieDataset getLeadStatusToCountDataset() {
		def statusCounts = new HashMultiset()
		issues.each {
			statusCounts.add(it.statusObject.name)
		}

		def dataset = new DefaultPieDataset()
		dataset.setValue("Won", statusCounts.count("Won"))
		dataset.setValue("Lost", statusCounts.count("Lost"))
		dataset.setValue("Open", statusCounts.count("Open"))

		return dataset
	}

	def getLeadStatusToCountCustomizer() {
		return { chart ->
			chart.plot.baseSectionOutlinePaint = RenderingUtils.BACKGROUND_COLOR
			chart.plot.setSectionPaint("Won", WON_COLOR)
			chart.plot.setSectionPaint("Lost", LOST_COLOR)
			chart.plot.setSectionPaint("Open", OPEN_COLOR)
		}
	}

	CategoryDataset getLeadStatusToValueDataset() {
		def issuesByStatus = new ArrayListMultimap()
		issues.each {
			issuesByStatus.put(it.statusObject.name, it)
		}

		def dataset = new DefaultCategoryDataset()
		dataset.setValue(getLeadTotalValue(issuesByStatus.get("Won")), "Won", "")
		dataset.setValue(getLeadTotalValue(issuesByStatus.get("Lost")), "Lost", "")
		dataset.setValue(getLeadTotalValue(issuesByStatus.get("Open")), "Open", "")

		return dataset
	}

	def getLeadStatusToValueCustomizer() {
		return { chart ->
			chart.plot.renderer.setSeriesPaint(0, WON_COLOR)
			chart.plot.renderer.setSeriesPaint(1, LOST_COLOR)
			chart.plot.renderer.setSeriesPaint(2, OPEN_COLOR)

			chart.plot.renderer.setSeriesOutlinePaint(0, RenderingUtils.BACKGROUND_COLOR)
			chart.plot.renderer.setSeriesOutlinePaint(1, RenderingUtils.BACKGROUND_COLOR)
			chart.plot.renderer.setSeriesOutlinePaint(2, RenderingUtils.BACKGROUND_COLOR)
		}
	}

	PieDataset getCountryToCountDataset() {
		def leadCountByCountry = new TreeMultiset()
		for(issue in issues) {
			if(isWonLead(issue)) {
				leadCountByCountry.add(getLeadCountry(issue))
			}
		}

		def dataset = new DefaultPieDataset()
		for(countryName in leadCountByCountry.elementSet()) {
			dataset.setValue(countryName, leadCountByCountry.count(countryName))
		}

		return dataset
	}

	def getCountryToCountCustomizer() {
		return { chart ->
			chart.plot.drawingSupplier = new GradientDrawingSupplier(chart.plot.dataset.itemCount)
			chart.plot.baseSectionOutlinePaint = RenderingUtils.BACKGROUND_COLOR
		}
	}

	CategoryDataset getCountryToValueDataset() {
		def leadsByCountry = TreeMultimap.create(Ordering.natural(), new IssueByUpdateDateComparator())
		for(issue in issues) {
			if(isWonLead(issue)) {
				leadsByCountry.put(getLeadCountry(issue), issue)
			}
		}

		def dataset = new DefaultCategoryDataset()
		for(countryName in leadsByCountry.keySet()) {
			def countrySign = StringUtils.isNotBlank(countryName) ? StringUtils.substring(countryName, 0, 2) : "n/a"
			dataset.setValue(getLeadTotalValue(leadsByCountry.get(countryName)), "value", countrySign)
		}

		return dataset
	}

	def getCountryToValueCustomizer() {
		return { chart ->
			chart.plot.renderer = new GradientBarRenderer(chart.categoryPlot.dataset.columnCount)
		}
	}

	// - Utils ---

	static def filterIssuesByStatusName(issues, statusName) {
		Predicate p = new Predicate() {
			boolean apply(issue) { return statusName == issue.statusObject.name.toLowerCase() }
		}

		return Collections2.filter(issues, p)
	}

	static String formatYearMonthDate(Date date) {
		return monthOnlyDateFormat.format(date)
	}

	static String formatMoneyValue(long moneyValue) {
		return moneyFormat.format(moneyValue)
	}

	private static String getPercentageString(value, divisor) {
		double p = (divisor != 0) ? value / divisor : 0
		return NumberFormat.getPercentInstance().format(p)
	}

	/**
	 * Returns the midnight moment in the first day of the month for the passed date.
	 */
	private static trimToMonth(date) {
		calendar.setTime(date)

		calendar.set(Calendar.DAY_OF_MONTH, 1)
		calendar.set(Calendar.HOUR_OF_DAY, 0)
		calendar.set(Calendar.MINUTE, 0)
		calendar.set(Calendar.SECOND, 0)
		calendar.set(Calendar.MILLISECOND, 0)

		return calendar.time
	}

	static class IssueByUpdateDateComparator implements Comparator {
		int compare(issue1,  issue2) { issue1.updated.compareTo(issue2.updated) }
	}
}
